//////////////////////////////////////////////////////////////////
// Symbolic
//
// Library for the construction of nested symbolic propositions
// The Flying Pig!
// Started 5/8/2003
// http://www.flypig.co.uk
// 
// 
// The MIT License
// See symbolic.h, COPYING file or website for licence
// 
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
// Includes

#include "symbolic.h"
#include "symbolic_private.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

//////////////////////////////////////////////////////////////////
// Defines

//////////////////////////////////////////////////////////////////
// Structures

//////////////////////////////////////////////////////////////////
// Global variables

#if defined _MEM_PROFILE
static int gnPropMemAllocated = 0;
#endif

//////////////////////////////////////////////////////////////////
// Function prototypes

bool        SubstituteRecursive     (Operation * psMain, Operation * psFind, Operation * psSub);
int         SubstituteRecursivePair (Operation * psMain, Operation * psFind1, Operation * psSub1, Operation * psFind2, Operation * psSub2);
int         CompareOperationsPair   (Operation * psMain, Operation * psCompare1, Operation * psCompare2);

//////////////////////////////////////////////////////////////////
// Main application

#if !defined _DEBUG
inline void noprintf (char * szNull, ...) {
  szNull = szNull;
}
#endif

//////////////////////////////////////////////////////////////////
// Reset memory allocation details
void PropMemReset (void) {
#if defined _MEM_PROFILE
  printf ("Memory profiling reset\n");
  gnPropMemAllocated = 0;
#else
  printf ("No memory profiling\n");
#endif
}

//////////////////////////////////////////////////////////////////
// Output current memory allocation details
void PropMemOutput (void) {
#if defined _MEM_PROFILE
  printf ("Total memory blocks allocated = %d\n", gnPropMemAllocated);
#endif
}

//////////////////////////////////////////////////////////////////
// void * malloc (size_t size)
void * PropMemMalloc (size_t size) {
#if defined _MEM_PROFILE
  gnPropMemAllocated += 1;
#endif
  return malloc (size);
}

//////////////////////////////////////////////////////////////////
// void * calloc (size_t n, size_t size)
void * PropMemCalloc (size_t n, size_t size) {
#if defined _MEM_PROFILE
  gnPropMemAllocated += 1;
#endif
  return calloc (n, size);
}

//////////////////////////////////////////////////////////////////
// void free (void * ptr)
void PropMemFree (void * ptr) {
#if defined _MEM_PROFILE
  gnPropMemAllocated -= 1;
#endif
  return free (ptr);
}

//////////////////////////////////////////////////////////////////
// Create a memory operation
Operation * CreateMemory (void) {
  Operation       *psOp;

  psOp = (Operation*)PropMalloc (sizeof(Operation));
  psOp->eOpType = OPTYPE_MEMORY;

  return psOp;
}

//////////////////////////////////////////////////////////////////
// Create an integer operation
Operation * CreateInteger (int const nInteger) {
  Operation       *psOp;

  psOp = (Operation*)PropMalloc (sizeof(Operation));
  psOp->eOpType = OPTYPE_INTEGER;
  psOp->Vars.nInteger = nInteger;

  return psOp;
}

//////////////////////////////////////////////////////////////////
// Create true/false truth value operation
Operation * CreateTruthValue (bool const boTruth) {
  Operation       *psOp;

  psOp = (Operation*)PropMalloc (sizeof(Operation));
  psOp->eOpType = OPTYPE_TRUTHVALUE;
  psOp->Vars.boTruth = boTruth;

  return psOp;
}

//////////////////////////////////////////////////////////////////
// Create a variable
Operation * CreateVariable (char const * szVar) {
  Operation       *psOp;

  psOp = (Operation*)PropMalloc (sizeof(Operation));
  psOp->eOpType = OPTYPE_VARIABLE;
  psOp->Vars.psVar = (OpVariable*)PropMalloc(sizeof(OpVariable));
  strncpy (psOp->Vars.psVar->szVar, szVar, MAXVARSIZE);
  psOp->Vars.psVar->psValue = NULL;

  return psOp;
}

//////////////////////////////////////////////////////////////////
// Create a unary operation
Operation * CreateUnary (OPUNARY eOpType, Operation * psVar1) {
  Operation       *psOp;

  psOp = (Operation*)PropMalloc (sizeof(Operation));
  psOp->eOpType = OPTYPE_UNARY;
  psOp->Vars.psUnary = (OpUnary*)PropMalloc(sizeof(OpUnary));
  psOp->Vars.psUnary->eOpType = eOpType;
  psOp->Vars.psUnary->psVar1 = psVar1;

  return psOp;
}

//////////////////////////////////////////////////////////////////
// Create a binary operation
Operation * CreateBinary (OPBINARY eOpType, Operation * psVar1, Operation * psVar2) {
  Operation       *psOp;

  psOp = (Operation*)PropMalloc (sizeof(Operation));
  psOp->eOpType = OPTYPE_BINARY;
  psOp->Vars.psBinary = (OpBinary*)PropMalloc(sizeof(OpBinary));
  psOp->Vars.psBinary->eOpType = eOpType;
  psOp->Vars.psBinary->psVar1 = psVar1;
  psOp->Vars.psBinary->psVar2 = psVar2;

  return psOp;
}

//////////////////////////////////////////////////////////////////
// Create a ternary operation
Operation * CreateTernary (OPTERNARY eOpType, Operation * psVar1, Operation * psVar2, Operation * psVar3) {
  Operation       *psOp;

  psOp = (Operation*)PropMalloc (sizeof(Operation));
  psOp->eOpType = OPTYPE_TERNARY;
  psOp->Vars.psTernary = (OpTernary*)PropMalloc(sizeof(OpTernary));
  psOp->Vars.psTernary->eOpType = eOpType;
  psOp->Vars.psTernary->psVar1 = psVar1;
  psOp->Vars.psTernary->psVar2 = psVar2;
  psOp->Vars.psTernary->psVar3 = psVar3;

  return psOp;
}

//////////////////////////////////////////////////////////////////
// Recursively free up all of the memory used by a formula and its sub formulas
void FreeRecursive (Operation * psOp) {
  if (psOp) {
    switch (psOp->eOpType) {
      case OPTYPE_MEMORY:
      case OPTYPE_INTEGER:
      case OPTYPE_TRUTHVALUE:
        // Nothing else to free - backtrack
        break;
      case OPTYPE_VARIABLE:
        // Free up the variable string and decrement
        // variable reference if there is one,
        // then backtrack
        if (psOp->Vars.psVar->psValue) {
          DecrementVarRef (psOp->Vars.psVar->psValue);
        }
        PropFree (psOp->Vars.psVar);
        break;
      case OPTYPE_UNARY:
        // Free up any operations further down the tree
        if (psOp->Vars.psUnary) {
          FreeRecursive (psOp->Vars.psUnary->psVar1);
          PropFree (psOp->Vars.psUnary);
        }
        // Then backtrack
        break;
      case OPTYPE_BINARY:
        // Free up any operations further down the tree
        if (psOp->Vars.psBinary) {
          FreeRecursive (psOp->Vars.psBinary->psVar1);
          FreeRecursive (psOp->Vars.psBinary->psVar2);
          PropFree (psOp->Vars.psBinary);
        }
        // Then backtrack
        break;
      case OPTYPE_TERNARY:
        // Free up any operations further down the tree
        if (psOp->Vars.psTernary) {
          FreeRecursive (psOp->Vars.psTernary->psVar1);
          FreeRecursive (psOp->Vars.psTernary->psVar2);
          FreeRecursive (psOp->Vars.psTernary->psVar3);
          PropFree (psOp->Vars.psTernary);
        }
        // Then backtrack
        break;
      default:
        printf("Invalid operation type\n");
        break;
    }

    PropFree (psOp);
  }
}

//////////////////////////////////////////////////////////////////
// Recursively copy a formula and all its subformulas
Operation * CopyRecursive (Operation * psOp)
{
  Operation         * psReturn = NULL;
  if (psOp) {
    switch (psOp->eOpType) {
      case OPTYPE_MEMORY:
        psReturn = CreateMemory ();
        break;
      case OPTYPE_INTEGER:
        psReturn = CreateInteger (psOp->Vars.nInteger);
        break;
      case OPTYPE_TRUTHVALUE:
        psReturn = CreateTruthValue (psOp->Vars.boTruth);
        break;
      case OPTYPE_VARIABLE:
        psReturn = CreateVariable (psOp->Vars.psVar->szVar);
        break;
      case OPTYPE_UNARY:
        psReturn = CreateUnary (psOp->Vars.psUnary->eOpType,
          CopyRecursive (psOp->Vars.psUnary->psVar1));
        break;
      case OPTYPE_BINARY:
        psReturn = CreateBinary (psOp->Vars.psBinary->eOpType,
          CopyRecursive (psOp->Vars.psBinary->psVar1),
          CopyRecursive (psOp->Vars.psBinary->psVar2));
        break;
      case OPTYPE_TERNARY:
        psReturn = CreateTernary (psOp->Vars.psTernary->eOpType,
          CopyRecursive (psOp->Vars.psTernary->psVar1),
          CopyRecursive (psOp->Vars.psTernary->psVar2),
          CopyRecursive (psOp->Vars.psTernary->psVar3));
        break;
      default:
        printf("Invalid operation type\n");
        break;
    }
  }
  return psReturn;
}

//////////////////////////////////////////////////////////////////
// Compare two formulae recursively
bool CompareOperations (Operation * psOp1, Operation * psOp2) {
  bool  boReturn = TRUE;

  if ((psOp1) && (psOp2)) {
    if (psOp1->eOpType == psOp2->eOpType) {
      switch (psOp1->eOpType) {
        case OPTYPE_MEMORY:
          // Return TRUE
          break;
        case OPTYPE_INTEGER:
          if (psOp1->Vars.nInteger != psOp2->Vars.nInteger) {
            boReturn = FALSE;
          }
          break;
        case OPTYPE_TRUTHVALUE:
          if (psOp1->Vars.boTruth != psOp2->Vars.boTruth) {
            boReturn = FALSE;
          }
          break;
        case OPTYPE_VARIABLE:
          if (strcmp (psOp1->Vars.psVar->szVar, psOp2->Vars.psVar->szVar) != 0) {
            boReturn = FALSE;
          }
          break;
        case OPTYPE_UNARY:
          if (psOp1->Vars.psUnary->eOpType != psOp2->Vars.psUnary->eOpType) {
            boReturn = FALSE;
          }
          else {
            boReturn = CompareOperations (psOp1->Vars.psUnary->psVar1,
              psOp2->Vars.psUnary->psVar1);
          }
          break;
        case OPTYPE_BINARY:
          if (psOp1->Vars.psBinary->eOpType != psOp2->Vars.psBinary->eOpType) {
            boReturn = FALSE;
          }
          else {
            boReturn = (CompareOperations (psOp1->Vars.psBinary->psVar1,
              psOp2->Vars.psBinary->psVar1)
              && CompareOperations (psOp1->Vars.psBinary->psVar2,
              psOp2->Vars.psBinary->psVar2));
          }
          break;
        case OPTYPE_TERNARY:
          if (psOp1->Vars.psTernary->eOpType != psOp2->Vars.psTernary->eOpType) {
            boReturn = FALSE;
          }
          else {
            boReturn = (CompareOperations (psOp1->Vars.psTernary->psVar1,
              psOp2->Vars.psTernary->psVar1)
              && CompareOperations (psOp1->Vars.psTernary->psVar2,
              psOp2->Vars.psTernary->psVar2)
              && CompareOperations (psOp1->Vars.psTernary->psVar3,
              psOp2->Vars.psTernary->psVar3));
          }
          break;
        default:
          printf("Invalid operation type\n");
          break;
      }
    }
    else {
      boReturn = FALSE;
    }
  }
  else {
    boReturn = FALSE;
  }

  return boReturn;
}

//////////////////////////////////////////////////////////////////
// Search a formula for a given subformula
Operation * FindOperation (Operation * psMain, Operation * psFind) {
  Operation       *psReturn = NULL;
  bool            boSame;

  if ((psMain) && (psFind)) {
    switch (psMain->eOpType) {
      case OPTYPE_MEMORY:
      case OPTYPE_VARIABLE:
      case OPTYPE_TRUTHVALUE:
      case OPTYPE_INTEGER:
        boSame = CompareOperations (psMain, psFind);
        if (boSame) {
          psReturn = psMain;
        }
        break;
      case OPTYPE_UNARY:
        psReturn = FindOperation (psMain->Vars.psUnary->psVar1, psFind);
        if (!psReturn) {
          boSame = CompareOperations (psMain, psFind);
          if (boSame) {
            psReturn = psMain;
          }
        }
        break;
      case OPTYPE_BINARY:
        psReturn = FindOperation (psMain->Vars.psBinary->psVar1, psFind);
        if (!psReturn) {
          psReturn = FindOperation (psMain->Vars.psBinary->psVar2, psFind);
          if (!psReturn) {
            boSame = CompareOperations (psMain, psFind);
            if (boSame) {
              psReturn = psMain;
            }
          }
        }
        break;
      case OPTYPE_TERNARY:
        psReturn = FindOperation (psMain->Vars.psTernary->psVar1, psFind);
        if (!psReturn) {
          psReturn = FindOperation (psMain->Vars.psTernary->psVar2, psFind);
          if (!psReturn) {
            psReturn = FindOperation (psMain->Vars.psTernary->psVar3, psFind);
            if (!psReturn) {
              boSame = CompareOperations (psMain, psFind);
              if (boSame) {
                psReturn = psMain;
              }
            }
          }
        }
        break;
      default:
        printf("Invalid operation type\n");
        break;
    }
  }
  return psReturn;
}

//////////////////////////////////////////////////////////////////
// Substitute all instances of a given subformula for a formula
Operation * SubstituteOperation (Operation * psMain, Operation * psFind, Operation * psSub) {
  bool          boFind;
  Operation     *psReturn;

  boFind = SubstituteRecursive (psMain, psFind, psSub);
  if (boFind) {
    FreeRecursive (psMain);
    psReturn = CopyRecursive (psSub);
  }
  else {
    psReturn = psMain;
  }
  return psReturn;
}

//////////////////////////////////////////////////////////////////
// Substitute recursively all instances of a given subformula for a formula
bool SubstituteRecursive (Operation * psMain, Operation * psFind, Operation * psSub) {
  bool            boSubstitute = FALSE;
  bool            boFind;

  if ((psMain) && (psSub)) {
    switch (psMain->eOpType) {
      case OPTYPE_MEMORY:
      case OPTYPE_VARIABLE:
      case OPTYPE_TRUTHVALUE:
      case OPTYPE_INTEGER:
        boSubstitute = CompareOperations (psMain, psFind);
        break;
      case OPTYPE_UNARY:
        boSubstitute = CompareOperations (psMain, psFind);
        if (!boSubstitute) {
          boFind = SubstituteRecursive (psMain->Vars.psUnary->psVar1, psFind, psSub);
          if (boFind) {
            FreeRecursive (psMain->Vars.psUnary->psVar1);
            psMain->Vars.psUnary->psVar1 = CopyRecursive (psSub);
          }
        }
        break;
      case OPTYPE_BINARY:
        boSubstitute = CompareOperations (psMain, psFind);
        if (!boSubstitute) {
          boFind = SubstituteRecursive (psMain->Vars.psBinary->psVar1,
            psFind, psSub);
          if (boFind) {
            FreeRecursive (psMain->Vars.psBinary->psVar1);
            psMain->Vars.psBinary->psVar1 = CopyRecursive (psSub);
          }
          boFind = SubstituteRecursive (psMain->Vars.psBinary->psVar2, psFind, psSub);
          if (boFind) {
            FreeRecursive (psMain->Vars.psBinary->psVar2);
            psMain->Vars.psBinary->psVar2 = CopyRecursive (psSub);
          }
        }
        break;
      case OPTYPE_TERNARY:
        boSubstitute = CompareOperations (psMain, psFind);
        if (!boSubstitute) {
          boFind = SubstituteRecursive (psMain->Vars.psTernary->psVar1, psFind, psSub);
          if (boFind) {
            FreeRecursive (psMain->Vars.psTernary->psVar1);
            psMain->Vars.psTernary->psVar1 = CopyRecursive (psSub);
          }
          boFind = SubstituteRecursive (psMain->Vars.psTernary->psVar2, psFind, psSub);
          if (boFind) {
            FreeRecursive (psMain->Vars.psTernary->psVar2);
            psMain->Vars.psTernary->psVar2 = CopyRecursive (psSub);
          }

          boFind = SubstituteRecursive (psMain->Vars.psTernary->psVar2, psFind, psSub);
          if (boFind) {
            FreeRecursive (psMain->Vars.psTernary->psVar2);
            psMain->Vars.psTernary->psVar2 = CopyRecursive (psSub);
          }
        }
        break;
      default:
        printf("Invalid operation type\n");
        break;
    }
  }
  return boSubstitute;
}

//////////////////////////////////////////////////////////////////
// Substitute all instances of a given pair of subformula for respective formulae
Operation * SubstituteOperationPair (Operation * psMain, Operation * psFind1, Operation * psSub1, Operation * psFind2, Operation * psSub2) {
  int           nFind;
  Operation     *psReturn;

  nFind = SubstituteRecursivePair (psMain, psFind1, psSub1, psFind2, psSub2);
  switch (nFind) {
    default:
    case 0:
      psReturn = psMain;
      break;
    case 1:
      FreeRecursive (psMain);
      psReturn = CopyRecursive (psSub1);
      break;
    case 2:
      FreeRecursive (psMain);
      psReturn = CopyRecursive (psSub2);
      break;
  }
  return psReturn;
}

//////////////////////////////////////////////////////////////////
// Substitute recursively all instances of a given pair of subformula for respective formulae
int SubstituteRecursivePair (Operation * psMain, Operation * psFind1, Operation * psSub1, Operation * psFind2, Operation * psSub2) {
  int             nSubstitute = 0;
  int             nFind;

  if ((psMain) && (psSub1) && (psSub2)) {
    switch (psMain->eOpType) {
      case OPTYPE_MEMORY:
      case OPTYPE_VARIABLE:
      case OPTYPE_TRUTHVALUE:
      case OPTYPE_INTEGER:
        nSubstitute = CompareOperationsPair (psMain, psFind1, psFind2);
        break;
      case OPTYPE_UNARY:
        nSubstitute = CompareOperationsPair (psMain, psFind1, psFind2);
        if (nSubstitute == 0) {
          nFind = SubstituteRecursivePair (psMain->Vars.psUnary->psVar1,
            psFind1, psSub1, psFind2, psSub2);
          switch (nFind) {
            case 1:
              FreeRecursive (psMain->Vars.psUnary->psVar1);
              psMain->Vars.psUnary->psVar1 = CopyRecursive (psSub1);
              break;
            case 2:
              FreeRecursive (psMain->Vars.psUnary->psVar1);
              psMain->Vars.psUnary->psVar1 = CopyRecursive (psSub2);
              break;
            default:
              // Do nothing
              break;
          }
        }
        break;
      case OPTYPE_BINARY:
        nSubstitute = CompareOperationsPair (psMain, psFind1, psFind2);
        if (nSubstitute == 0) {
          nFind = SubstituteRecursivePair (psMain->Vars.psBinary->psVar1,
            psFind1, psSub1, psFind2, psSub2);
          switch (nFind) {
            case 1:
              FreeRecursive (psMain->Vars.psBinary->psVar1);
              psMain->Vars.psBinary->psVar1 = CopyRecursive (psSub1);
              break;
            case 2:
              FreeRecursive (psMain->Vars.psBinary->psVar1);
              psMain->Vars.psBinary->psVar1 = CopyRecursive (psSub2);
              break;
            default:
              // Do nothing
              break;
          }
          nFind = SubstituteRecursivePair (psMain->Vars.psBinary->psVar2,
            psFind1, psSub1, psFind2, psSub2);
          switch (nFind) {
            case 1:
              FreeRecursive (psMain->Vars.psBinary->psVar2);
              psMain->Vars.psBinary->psVar2 = CopyRecursive (psSub1);
              break;
            case 2:
              FreeRecursive (psMain->Vars.psBinary->psVar2);
              psMain->Vars.psBinary->psVar2 = CopyRecursive (psSub2);
              break;
            default:
              // Do nothing
              break;
          }
        }
        break;
      case OPTYPE_TERNARY:
        nSubstitute = CompareOperationsPair (psMain, psFind1, psFind2);
        if (nSubstitute == 0) {
          nFind = SubstituteRecursivePair (psMain->Vars.psTernary->psVar1,
            psFind1, psSub1, psFind2, psSub2);
          switch (nFind) {
            case 1:
              FreeRecursive (psMain->Vars.psTernary->psVar1);
              psMain->Vars.psTernary->psVar1 = CopyRecursive (psSub1);
              break;
            case 2:
              FreeRecursive (psMain->Vars.psTernary->psVar1);
              psMain->Vars.psTernary->psVar1 = CopyRecursive (psSub2);
              break;
            default:
              // Do nothing
              break;
          }
          nFind = SubstituteRecursivePair (psMain->Vars.psTernary->psVar2, psFind1, psSub1, psFind2, psSub2);
          switch (nFind) {
            case 1:
              FreeRecursive (psMain->Vars.psTernary->psVar2);
              psMain->Vars.psTernary->psVar2 = CopyRecursive (psSub1);
              break;
            case 2:
              FreeRecursive (psMain->Vars.psTernary->psVar2);
              psMain->Vars.psTernary->psVar2 = CopyRecursive (psSub2);
              break;
            default:
              // Do nothing
              break;
          }
          nFind = SubstituteRecursivePair (psMain->Vars.psTernary->psVar3, psFind1, psSub1, psFind2, psSub2);
          switch (nFind) {
            case 1:
              FreeRecursive (psMain->Vars.psTernary->psVar3);
              psMain->Vars.psTernary->psVar3 = CopyRecursive (psSub1);
              break;
            case 2:
              FreeRecursive (psMain->Vars.psTernary->psVar3);
              psMain->Vars.psTernary->psVar3 = CopyRecursive (psSub2);
              break;
            default:
              // Do nothing
              break;
          }
        }
        break;
      default:
        printf("Invalid operation type\n");
        break;
    }
  }
  return nSubstitute;
}

//////////////////////////////////////////////////////////////////
// Compare a pair of formulae against a formula recursively
int CompareOperationsPair (Operation * psMain, Operation * psCompare1, Operation * psCompare2) {
  int       nReturn = 0;

  if (CompareOperations (psMain, psCompare1)) {
    nReturn = 1;
  }
  else {
    if (CompareOperations (psMain, psCompare2)) {
      nReturn = 2;
    }
  }
  return nReturn;
}

